import std/json, std/options, std/sequtils

import ../private/jsonutils

type
  JwkKeyType* = enum EcPublic, EcPrivate, RsaPublic, RsaPrivate, Octet
  JwkKty* = enum ktyEC = "EC", ktyRSA = "RSA", ktyOctet = "oct"

  JwkUse* = enum jwkuSignature = "sig", jwkuEncrypt = "enc"

  JwkOps* = enum jwkopSign = "sign", jwkopVerify = "verify",
                 jwkopEncrypt = "encrypt", jwkopDecrypt = "decrypt",
                 jwkopWrapKey = "wrapKey", jwkopUnwrapKey = "unwrapKey",
                 jwkopDeriveKey = "deriveKey", jwkopDeriveBits = "deriveBits"

  JwkAlg* = enum
    HS256 = "HS256", HS384 = "HS384", HS512 = "HS512",
    RS256 = "RS256", RS384 = "RS384", RS512 = "RS512",
    ES256 = "ES256", ES384 = "ES384", ES512 = "ES512",
    PS256 = "PS256", PS384 = "PS384", PS512 = "PS512",

    algNone = "none", algDir = "dir",

    RSA1_5 = "RSA1_5", RSA_OAEP = "RSA-OAEP", RSA_OAEP_256 = "RSA-OEAP-256",

    A128KW = "A128KW", A192KW = "A192KW", A256KW = "A256KW",

    ECDH_ES = "ECDH_ES", ECDH_ES_A128KW = "ECDH_ES_A128KW",
    ECDH_ES_A192KW = "ECDH_ES_A192KW", ECDH_ES_A256KW = "ECDH_ES_A256KW",

    A128GCMKW = "A128GCMKW", A192GCMKW = "A192GCMKW", A256GCMKW = "A256GCMKW",

    PBES2_HS256_A128KW = "PBES2_HS256_A128KW",
    PBES2_HS384_A192KW = "PBES2_HS384_A192KW",
    PBES2_HS512_A256KW = "PBES2_HS512_A256KW",

    A128CBC_HS256 = "A128CBC_HS256",
    A192CBC_HS384 = "A192CBC_HS384",
    A256CBC_HS512 = "A256CBC_HS512",

    A128GCM = "A128GCM", A192GCM = "A192GCM", A256GCM = "A256GCM"

  EcCrv* = enum P256 = "P-256", P384 = "P-384", P521 = "P-521"

  EcPubKey = object of RootObj
    crv*: string
    x*: string
    y*: Option[string]

  EcPrvKey* = object of EcPubKey
    d*: string

  RsaPubKey* = object of RootObj
    n*: string
    e*: string

  RsaOtherPrime* = object
    r*: string
    d*: string
    t*: string

  RsaPrvKey* = object of RsaPubKey
    d*: string
    p*: Option[string]
    q*: Option[string]
    dp*: Option[string]
    dq*: Option[string]
    qi*: Option[string]
    oth*: Option[seq[RsaOtherPrime]]

  OctetKey* = object
    k*: string

  JWK* = object
    json: JsonNode
    kty*: string
    use*: Option[string]
    key_ops*: Option[string]
    alg*: Option[string]
    kid*: Option[string]
    x5u*: Option[string]
    x5c*: Option[string]
    x5t*: Option[string]
    x5tS256*: Option[string]

    case keyKind*: JwkKeyType
    of EcPublic: ecPub*: EcPubKey
    of EcPrivate: ecPrv*: EcPrvKey
    of RsaPublic: rsaPub*: RsaPubKey
    of RsaPrivate: rsaPrv*: RsaPrvKey
    of Octet: octKey*: OctetKey

  JwkSet* = seq[JWK]

func `[]`*(jwk: JWK, key: string): JsonNode = jwk.json[key]
func `[]=`*(jwk: JWK, key: string, val: JsonNode): void = jwk.json[key] = val

func parseEcPubKey*(n: JsonNode): EcPubKey =
  EcPubKey(
    crv: n.reqStrVal("crv"),
    x: n.reqStrVal("x"),
    y: n.optStrVal("y"))

func parseEcPrvKey*(n: JsonNode): EcPrvKey =
  let pk = parseEcPubKey(n)
  EcPrvKey(crv: pk.crv, x: pk.x, y: pk.y, d: n.reqStrVal("d"))

func parseRsaPubKey*(n: JsonNode): RsaPubKey =
  RsaPubKey(
    n: n.reqStrVal("n"),
    e: n.reqStrVal("e"))

func parseRsaOtherPrime(n: JsonNode): RsaOtherPrime =
  RsaOtherPrime(
    r: n.reqStrVal("r"),
    d: n.reqStrVal("d"),
    t: n.reqStrVal("t"))

func parseRsaPrvKey*(n: JsonNode): RsaPrvKey =
  let pk = parseRsaPubKey(n)
  RsaPrvKey(n: pk.n, e: pk.e,
    d: n.reqStrVal("d"),
    p: n.optStrVal("p"),
    q: n.optStrVal("q"),
    dp: n.optStrVal("dp"),
    dq: n.optStrVal("dq"),
    qi: n.optStrVal("qi"),
    oth:
      if n.hasKey("oth"):
        some(n["oth"].getElems.mapIt(parseRsaOtherPrime(it)))
      else: none[seq[RsaOtherPrime]]()
  )


func parseOctetKey*(n: JsonNode): OctetKey =
  OctetKey(k: n.reqStrVal("k"))

func parseJwk*(n: JsonNode): JWK =
  # TODO: documentation, example, handle encrypted keys (JWEs)
  case n["kty"].getStr(""):

  of $ktyEC:
    if n.hasKey("d"):
      result = JWK(
        kty: $ktyEC,
        keyKind: EcPrivate,
        ecPrv: parseEcPrvKey(n))
    else:
      result = JWK(
        kty: $ktyEC,
        keyKind: EcPublic,
        ecPub: parseEcPubKey(n))

  of $ktyRSA:
    if n.hasKey("d"):
      result = JWK(
        kty: $ktyRSA,
        keyKind: RsaPrivate,
        rsaPrv: parseRsaPrvKey(n))
    else:
      result = JWK(
        kty: $ktyRSA,
        keyKind: RsaPublic,
        rsaPub: parseRsaPubKey(n))

  of $ktyOctet:
    result = JWK(
      kty: $ktyOctet,
      keyKind: Octet,
      octKey: parseOctetKey(n))

  else: raise newException(ValueError,
    "Unrecognized or missing kty: '" & n["kty"].getStr("") & "'")

  result.json = n
  result.use = n.optStrVal("use")
  result.key_ops = n.optStrVal("key_ops")
  result.alg = n.optStrVal("alg")
  result.kid = n.optStrVal("kid")
  result.x5u = n.optStrVal("x5u")
  result.x5c = n.optStrVal("x5c")
  result.x5t = n.optStrVal("x5t")
  result.x5tS256 = n.optStrVal("x5t#S256")

func parseJwkSet*(n: JsonNode): JwkSet =
  # TODO: documentation, examples, handled encrypted set (JWE)
  if not n.hasKey("keys") or n["keys"].kind != JArray:
    raise newException(ValueError, "JWK Set is missing the 'keys' member, " &
      "or 'keys' is not an array.")

  return n["keys"].getElems.mapIt(parseJwk(it))
